@@ -16,10 +16,10 @@ import (
 	"time"
 
 	"golang.org/x/tools/go/packages"
+	"golang.org/x/tools/gopls/internal/bug"
 	"golang.org/x/tools/gopls/internal/lsp/protocol"
 	"golang.org/x/tools/gopls/internal/lsp/source"
 	"golang.org/x/tools/gopls/internal/span"
-	"golang.org/x/tools/internal/bug"
 	"golang.org/x/tools/internal/event"
 	"golang.org/x/tools/internal/event/tag"
 	"golang.org/x/tools/internal/gocommand"
@@ -61,7 +61,7 @@ func (s *snapshot) load(ctx context.Context, allowNetwork bool, scopes ...loadSc
 				// Don't try to load a file that doesn't exist, or isn't a go file.
 				continue
 			}
-			contents, err := fh.Read()
+			contents, err := fh.Content()
 			if err != nil {
 				continue
 			}
@@ -72,14 +72,9 @@ func (s *snapshot) load(ctx context.Context, allowNetwork bool, scopes ...loadSc
 			}
 
 		case moduleLoadScope:
-			switch scope {
-			case "std", "cmd":
-				query = append(query, string(scope))
-			default:
-				modQuery := fmt.Sprintf("%s/...", scope)
-				query = append(query, modQuery)
-				moduleQueries[modQuery] = string(scope)
-			}
+			modQuery := fmt.Sprintf("%s%c...", scope.dir, filepath.Separator)
+			query = append(query, modQuery)
+			moduleQueries[modQuery] = string(scope.modulePath)
 
 		case viewLoadScope:
 			// If we are outside of GOPATH, a module, or some other known
@@ -103,7 +98,7 @@ func (s *snapshot) load(ctx context.Context, allowNetwork bool, scopes ...loadSc
 	}
 	sort.Strings(query) // for determinism
 
-	ctx, done := event.Start(ctx, "cache.view.load", tag.Query.Of(query))
+	ctx, done := event.Start(ctx, "cache.snapshot.load", tag.Query.Of(query))
 	defer done()
 
 	flags := source.LoadWorkspace
@@ -149,6 +144,31 @@ func (s *snapshot) load(ctx context.Context, allowNetwork bool, scopes ...loadSc
 		return fmt.Errorf("packages.Load error: %w", err)
 	}
 
+	// Workaround for a bug (?) that has been in go/packages since
+	// the outset: Package("unsafe").GoFiles=[], whereas it should
+	// include unsafe/unsafe.go. Derive it from builtins.go.
+	//
+	// This workaround relies on the fact that we always add both
+	// builtins and unsafe to the set of scopes in the workspace load.
+	//
+	// TODO(adonovan): fix upstream in go/packages.
+	// (Does this need a proposal? Arguably not.)
+	{
+		var builtin, unsafe *packages.Package
+		for _, pkg := range pkgs {
+			if pkg.ID == "unsafe" {
+				unsafe = pkg
+			} else if pkg.ID == "builtin" {
+				builtin = pkg
+			}
+		}
+		if builtin != nil && unsafe != nil && len(builtin.GoFiles) == 1 {
+			unsafe.GoFiles = []string{
+				filepath.Join(filepath.Dir(builtin.GoFiles[0]), "../unsafe/unsafe.go"),
+			}
+		}
+	}
+
 	moduleErrs := make(map[string][]packages.Error) // module path -> errors
 	filterFunc := s.view.filterFunc()
 	newMetadata := make(map[PackageID]*source.Metadata)
@@ -201,31 +221,37 @@ func (s *snapshot) load(ctx context.Context, allowNetwork bool, scopes ...loadSc
 		if allFilesExcluded(pkg.GoFiles, filterFunc) {
 			continue
 		}
-		if err := buildMetadata(ctx, pkg, cfg, query, newMetadata, nil); err != nil {
-			return err
-		}
+		buildMetadata(newMetadata, pkg, cfg.Dir, query)
 	}
 
 	s.mu.Lock()
 
+	// Assert the invariant s.packages.Get(id).m == s.meta.metadata[id].
+	s.packages.Range(func(k, v interface{}) {
+		id, ph := k.(PackageID), v.(*packageHandle)
+		if s.meta.metadata[id] != ph.m {
+			panic("inconsistent metadata")
+		}
+	})
+
 	// Compute the minimal metadata updates (for Clone)
-	// required to preserve this invariant:
-	// for all id, s.packages.Get(id).m == s.meta.metadata[id].
+	// required to preserve the above invariant.
+	var files []span.URI // files to preload
+	seenFiles := make(map[span.URI]bool)
 	updates := make(map[PackageID]*source.Metadata)
 	for _, m := range newMetadata {
 		if existing := s.meta.metadata[m.ID]; existing == nil {
+			// Record any new files we should pre-load.
+			for _, uri := range m.CompiledGoFiles {
+				if !seenFiles[uri] {
+					seenFiles[uri] = true
+					files = append(files, uri)
+				}
+			}
 			updates[m.ID] = m
 			delete(s.shouldLoad, m.ID)
 		}
 	}
-	// Assert the invariant.
-	s.packages.Range(func(k, v interface{}) {
-		id, ph := k.(PackageID), v.(*packageHandle)
-		if s.meta.metadata[id] != ph.m {
-			// TODO(adonovan): upgrade to unconditional panic after Jan 2023.
-			bug.Reportf("inconsistent metadata")
-		}
-	})
 
 	event.Log(ctx, fmt.Sprintf("%s: updating metadata for %d packages", eventName, len(updates)))
 
@@ -235,31 +261,26 @@ func (s *snapshot) load(ctx context.Context, allowNetwork bool, scopes ...loadSc
 	meta := s.meta.Clone(updates)
 	workspacePackages := computeWorkspacePackagesLocked(s, meta)
 	for _, update := range updates {
-		if err := computeLoadDiagnostics(ctx, update, meta, lockedSnapshot{s}, workspacePackages); err != nil {
-			return err
-		}
+		computeLoadDiagnostics(ctx, update, meta, lockedSnapshot{s}, workspacePackages)
 	}
 	s.meta = meta
 	s.workspacePackages = workspacePackages
 	s.resetActivePackagesLocked()
 
-	s.dumpWorkspace("load")
 	s.mu.Unlock()
 
-	// Recompute the workspace package handle for any packages we invalidated.
+	// Opt: preLoad files in parallel.
 	//
-	// This is (putatively) an optimization since handle construction prefetches
-	// the content of all Go source files.
-	//
-	// However, one necessary side effect of this operation is that we are
-	// guaranteed to visit all package files during load. This is required for
-	// e.g. determining the set of directories to watch.
+	// Requesting files in batch optimizes the underlying filesystem reads.
+	// However, this is also currently necessary for correctness: populating all
+	// files in the snapshot is necessary for certain operations that rely on the
+	// completeness of the file map, e.g. computing the set of directories to
+	// watch.
 	//
 	// TODO(rfindley, golang/go#57558): determine the set of directories based on
-	// loaded packages, and skip this precomputation.
-	for _, m := range updates {
-		s.buildPackageHandle(ctx, m.ID) // ignore error
-	}
+	// loaded packages, so that reading files here is not necessary for
+	// correctness.
+	s.preloadFiles(ctx, files)
 
 	if len(moduleErrs) > 0 {
 		return &moduleErrorMap{moduleErrs}
@@ -425,35 +446,31 @@ func (s *snapshot) applyCriticalErrorToFiles(ctx context.Context, msg string, fi
 // buildMetadata populates the updates map with metadata updates to
 // apply, based on the given pkg. It recurs through pkg.Imports to ensure that
 // metadata exists for all dependencies.
-func buildMetadata(ctx context.Context, pkg *packages.Package, cfg *packages.Config, query []string, updates map[PackageID]*source.Metadata, path []PackageID) error {
+func buildMetadata(updates map[PackageID]*source.Metadata, pkg *packages.Package, loadDir string, query []string) {
 	// Allow for multiple ad-hoc packages in the workspace (see #47584).
 	pkgPath := PackagePath(pkg.PkgPath)
 	id := PackageID(pkg.ID)
+
+	// TODO(rfindley): this creates at most one command-line-arguments package
+	// per load, but if we pass multiple file= queries to go/packages, there may
+	// be multiple command-line-arguments packages.
+	//
+	// As reported in golang/go#59318, this can result in accidentally quadratic
+	// loading behavior.
 	if source.IsCommandLineArguments(id) {
 		suffix := ":" + strings.Join(query, ",")
 		id = PackageID(pkg.ID + suffix)
 		pkgPath = PackagePath(pkg.PkgPath + suffix)
 	}
 
+	// Duplicate?
 	if _, ok := updates[id]; ok {
-		// If we've already seen this dependency, there may be an import cycle, or
-		// we may have reached the same package transitively via distinct paths.
-		// Check the path to confirm.
-
-		// TODO(rfindley): this doesn't look sufficient. Any single piece of new
-		// metadata could theoretically introduce import cycles in the metadata
-		// graph. What's the point of this limited check here (and is it even
-		// possible to get an import cycle in data from go/packages)? Consider
-		// simply returning, so that this function need not return an error.
-		//
-		// We should consider doing a more complete guard against import cycles
-		// elsewhere.
-		for _, prev := range path {
-			if prev == id {
-				return fmt.Errorf("import cycle detected: %q", id)
-			}
-		}
-		return nil
+		// A package was encountered twice due to shared
+		// subgraphs (common) or cycles (rare). Although "go
+		// list" usually breaks cycles, we don't rely on it.
+		// breakImportCycles in metadataGraph.Clone takes care
+		// of it later.
+		return
 	}
 
 	// Recreate the metadata rather than reusing it to avoid locking.
@@ -463,7 +480,7 @@ func buildMetadata(ctx context.Context, pkg *packages.Package, cfg *packages.Con
 		Name:       PackageName(pkg.Name),
 		ForTest:    PackagePath(packagesinternal.GetForTest(pkg)),
 		TypesSizes: pkg.TypesSizes,
-		LoadDir:    cfg.Dir,
+		LoadDir:    loadDir,
 		Module:     pkg.Module,
 		Errors:     pkg.Errors,
 		DepsErrors: packagesinternal.GetDepsErrors(pkg),
@@ -556,25 +573,30 @@ func buildMetadata(ctx context.Context, pkg *packages.Package, cfg *packages.Con
 			continue
 		}
 
+		// Don't record self-import edges.
+		// (This simplifies metadataGraph's cycle check.)
+		if PackageID(imported.ID) == id {
+			if len(pkg.Errors) == 0 {
+				bug.Reportf("self-import without error in package %s", id)
+			}
+			continue
+		}
+
 		depsByImpPath[importPath] = PackageID(imported.ID)
 		depsByPkgPath[PackagePath(imported.PkgPath)] = PackageID(imported.ID)
-		if err := buildMetadata(ctx, imported, cfg, query, updates, append(path, id)); err != nil {
-			event.Error(ctx, "error in dependency", err)
-		}
+		buildMetadata(updates, imported, loadDir, query)
 	}
 	m.DepsByImpPath = depsByImpPath
 	m.DepsByPkgPath = depsByPkgPath
 
 	// m.Diagnostics is set later in the loading pass, using
 	// computeLoadDiagnostics.
-
-	return nil
 }
 
 // computeLoadDiagnostics computes and sets m.Diagnostics for the given metadata m.
 //
 // It should only be called during metadata construction in snapshot.load.
-func computeLoadDiagnostics(ctx context.Context, m *source.Metadata, meta *metadataGraph, fs source.FileSource, workspacePackages map[PackageID]PackagePath) error {
+func computeLoadDiagnostics(ctx context.Context, m *source.Metadata, meta *metadataGraph, fs source.FileSource, workspacePackages map[PackageID]PackagePath) {
 	for _, packagesErr := range m.Errors {
 		// Filter out parse errors from go list. We'll get them when we
 		// actually parse, and buggy overlay support may generate spurious
@@ -602,10 +624,8 @@ func computeLoadDiagnostics(ctx context.Context, m *source.Metadata, meta *metad
 			// not normally fail.
 			event.Error(ctx, "unable to compute deps errors", err, tag.Package.Of(string(m.ID)))
 		}
-		return nil
 	}
 	m.Diagnostics = append(m.Diagnostics, depsDiags...)
-	return nil
 }
 
 // containsPackageLocked reports whether p is a workspace package for the
@@ -700,8 +720,9 @@ func containsFileInWorkspaceLocked(s *snapshot, m *source.Metadata) bool {
 	return false
 }
 
-// computeWorkspacePackagesLocked computes workspace packages in the snapshot s
-// for the given metadata graph.
+// computeWorkspacePackagesLocked computes workspace packages in the
+// snapshot s for the given metadata graph. The result does not
+// contain intermediate test variants.
 //
 // s.mu must be held while calling this function.
 func computeWorkspacePackagesLocked(s *snapshot, meta *metadataGraph) map[PackageID]PackagePath {
@@ -735,6 +756,7 @@ func computeWorkspacePackagesLocked(s *snapshot, meta *metadataGraph) map[Packag
 			//
 			// Notably, this excludes intermediate test variants from workspace
 			// packages.
+			assert(!m.IsIntermediateTestVariant(), "unexpected ITV")
 			workspacePackages[m.ID] = m.ForTest
 		}
 	}
