@@ -29,12 +29,14 @@ import (
 	"golang.org/x/sync/errgroup"
 	"golang.org/x/tools/go/packages"
 	"golang.org/x/tools/go/types/objectpath"
+	"golang.org/x/tools/gopls/internal/bug"
+	"golang.org/x/tools/gopls/internal/lsp/filecache"
 	"golang.org/x/tools/gopls/internal/lsp/protocol"
 	"golang.org/x/tools/gopls/internal/lsp/source"
 	"golang.org/x/tools/gopls/internal/lsp/source/methodsets"
+	"golang.org/x/tools/gopls/internal/lsp/source/typerefs"
 	"golang.org/x/tools/gopls/internal/lsp/source/xrefs"
 	"golang.org/x/tools/gopls/internal/span"
-	"golang.org/x/tools/internal/bug"
 	"golang.org/x/tools/internal/event"
 	"golang.org/x/tools/internal/event/tag"
 	"golang.org/x/tools/internal/gocommand"
@@ -72,14 +74,17 @@ type snapshot struct {
 	// view.initializationSema.
 	initialized bool
 	// initializedErr holds the last error resulting from initialization. If
-	// initialization fails, we only retry when the the workspace modules change,
+	// initialization fails, we only retry when the workspace modules change,
 	// to avoid too many go/packages calls.
 	initializedErr *source.CriticalError
 
 	// mu guards all of the maps in the snapshot, as well as the builtin URI.
 	mu sync.Mutex
 
-	// builtin pins the AST and package for builtin.go in memory.
+	// builtin is the location of builtin.go in GOROOT.
+	//
+	// TODO(rfindley): would it make more sense to eagerly parse builtin, and
+	// instead store a *ParsedGoFile here?
 	builtin span.URI
 
 	// meta holds loaded metadata.
@@ -121,7 +126,7 @@ type snapshot struct {
 	analyses *persistent.Map // from analysisKey to analysisPromise
 
 	// workspacePackages contains the workspace's packages, which are loaded
-	// when the view is created.
+	// when the view is created. It contains no intermediate test variants.
 	workspacePackages map[PackageID]PackagePath
 
 	// shouldLoad tracks packages that need to be reloaded, mapping a PackageID
@@ -169,6 +174,17 @@ type snapshot struct {
 	// This set is immutable inside the snapshot, and therefore is not guarded by mu.
 	workspaceModFiles    map[span.URI]struct{}
 	workspaceModFilesErr error // error encountered computing workspaceModFiles
+
+	// importGraph holds a shared import graph to use for type-checking. Adding
+	// more packages to this import graph can speed up type checking, at the
+	// expense of in-use memory.
+	//
+	// See getImportGraph for additional documentation.
+	importGraphDone chan struct{} // closed when importGraph is set; may be nil
+	importGraph     *importGraph  // copied from preceding snapshot and re-evaluated
+
+	// pkgIndex is an index of package IDs, for efficient storage of typerefs.
+	pkgIndex *typerefs.PackageIndex
 }
 
 var globalSnapshotID uint64
@@ -213,7 +229,7 @@ func (s *snapshot) awaitPromise(ctx context.Context, p *memoize.Promise) (interf
 // The destroyedBy argument is used for debugging.
 //
 // v.snapshotMu must be held while calling this function, in order to preserve
-// the invariants described by the the docstring for v.snapshot.
+// the invariants described by the docstring for v.snapshot.
 func (v *View) destroy(s *snapshot, destroyedBy string) {
 	v.snapshotWG.Add(1)
 	go func() {
@@ -531,11 +547,11 @@ func (s *snapshot) goCommandInvocation(ctx context.Context, flags source.Invocat
 
 	var modContent []byte
 	if modURI != "" {
-		modFH, err := s.GetFile(ctx, modURI)
+		modFH, err := s.ReadFile(ctx, modURI)
 		if err != nil {
 			return "", nil, cleanup, err
 		}
-		modContent, err = modFH.Read()
+		modContent, err = modFH.Content()
 		if err != nil {
 			return "", nil, cleanup, err
 		}
@@ -588,7 +604,7 @@ func (s *snapshot) goCommandInvocation(ctx context.Context, flags source.Invocat
 		if modURI == "" {
 			return "", nil, cleanup, fmt.Errorf("no go.mod file found in %s", inv.WorkingDir)
 		}
-		modFH, err := s.GetFile(ctx, modURI)
+		modFH, err := s.ReadFile(ctx, modURI)
 		if err != nil {
 			return "", nil, cleanup, err
 		}
@@ -605,20 +621,32 @@ func (s *snapshot) goCommandInvocation(ctx context.Context, flags source.Invocat
 }
 
 func (s *snapshot) buildOverlay() map[string][]byte {
+	overlays := make(map[string][]byte)
+	for _, overlay := range s.overlays() {
+		if overlay.saved {
+			continue
+		}
+		// TODO(rfindley): previously, there was a todo here to make sure we don't
+		// send overlays outside of the current view. IMO we should instead make
+		// sure this doesn't matter.
+		overlays[overlay.URI().Filename()] = overlay.content
+	}
+	return overlays
+}
+
+// TODO(rfindley): investigate whether it would be worthwhile to keep track of
+// overlays when we get them via GetFile.
+func (s *snapshot) overlays() []*Overlay {
 	s.mu.Lock()
 	defer s.mu.Unlock()
 
-	overlays := make(map[string][]byte)
+	var overlays []*Overlay
 	s.files.Range(func(uri span.URI, fh source.FileHandle) {
 		overlay, ok := fh.(*Overlay)
 		if !ok {
 			return
 		}
-		if overlay.saved {
-			return
-		}
-		// TODO(rstambler): Make sure not to send overlays outside of the current view.
-		overlays[uri.Filename()] = overlay.content
+		overlays = append(overlays, overlay)
 	})
 	return overlays
 }
@@ -630,37 +658,59 @@ const (
 	methodSetsKind  = "methodsets"
 	exportDataKind  = "export"
 	diagnosticsKind = "diagnostics"
+	typerefsKind    = "typerefs"
 )
 
 func (s *snapshot) PackageDiagnostics(ctx context.Context, ids ...PackageID) (map[span.URI][]*source.Diagnostic, error) {
-	// TODO(rfindley): opt: avoid unnecessary encode->decode after type-checking.
-	data, err := s.getPackageData(ctx, diagnosticsKind, ids, func(p *syntaxPackage) []byte {
-		return encodeDiagnostics(p.diagnostics)
-	})
+	ctx, done := event.Start(ctx, "cache.snapshot.PackageDiagnostics")
+	defer done()
+
+	var mu sync.Mutex
 	perFile := make(map[span.URI][]*source.Diagnostic)
-	for _, data := range data {
-		if data != nil {
-			for _, diag := range data.m.Diagnostics {
-				perFile[diag.URI] = append(perFile[diag.URI], diag)
-			}
-			diags := decodeDiagnostics(data.data)
-			for _, diag := range diags {
-				perFile[diag.URI] = append(perFile[diag.URI], diag)
-			}
+	collect := func(diags []*source.Diagnostic) {
+		mu.Lock()
+		defer mu.Unlock()
+		for _, diag := range diags {
+			perFile[diag.URI] = append(perFile[diag.URI], diag)
 		}
 	}
-	return perFile, err
+	pre := func(i int, ph *packageHandle) bool {
+		data, err := filecache.Get(diagnosticsKind, ph.key)
+		if err == nil { // hit
+			collect(ph.m.Diagnostics)
+			collect(decodeDiagnostics(data))
+			return false
+		} else if err != filecache.ErrNotFound {
+			event.Error(ctx, "reading diagnostics from filecache", err)
+		}
+		return true
+	}
+	post := func(_ int, pkg *Package) {
+		collect(pkg.ph.m.Diagnostics)
+		collect(pkg.pkg.diagnostics)
+	}
+	return perFile, s.forEachPackage(ctx, ids, pre, post)
 }
 
 func (s *snapshot) References(ctx context.Context, ids ...PackageID) ([]source.XrefIndex, error) {
-	data, err := s.getPackageData(ctx, xrefsKind, ids, func(p *syntaxPackage) []byte { return p.xrefs })
+	ctx, done := event.Start(ctx, "cache.snapshot.References")
+	defer done()
+
 	indexes := make([]source.XrefIndex, len(ids))
-	for i, data := range data {
-		if data != nil {
-			indexes[i] = XrefIndex{m: data.m, data: data.data}
+	pre := func(i int, ph *packageHandle) bool {
+		data, err := filecache.Get(xrefsKind, ph.key)
+		if err == nil { // hit
+			indexes[i] = XrefIndex{m: ph.m, data: data}
+			return false
+		} else if err != filecache.ErrNotFound {
+			event.Error(ctx, "reading xrefs from filecache", err)
 		}
+		return true
 	}
-	return indexes, err
+	post := func(i int, pkg *Package) {
+		indexes[i] = XrefIndex{m: pkg.ph.m, data: pkg.pkg.xrefs}
+	}
+	return indexes, s.forEachPackage(ctx, ids, pre, post)
 }
 
 // An XrefIndex is a helper for looking up a package in a given package.
@@ -674,21 +724,24 @@ func (index XrefIndex) Lookup(targets map[PackagePath]map[objectpath.Path]struct
 }
 
 func (s *snapshot) MethodSets(ctx context.Context, ids ...PackageID) ([]*methodsets.Index, error) {
-	// TODO(rfindley): opt: avoid unnecessary encode->decode after type-checking.
-	data, err := s.getPackageData(ctx, methodSetsKind, ids, func(p *syntaxPackage) []byte {
-		return p.methodsets.Encode()
-	})
+	ctx, done := event.Start(ctx, "cache.snapshot.MethodSets")
+	defer done()
+
 	indexes := make([]*methodsets.Index, len(ids))
-	for i, data := range data {
-		if data != nil {
-			indexes[i] = methodsets.Decode(data.data)
-		} else if ids[i] == "unsafe" {
-			indexes[i] = &methodsets.Index{}
-		} else {
-			panic(fmt.Sprintf("nil data for %s", ids[i]))
+	pre := func(i int, ph *packageHandle) bool {
+		data, err := filecache.Get(methodSetsKind, ph.key)
+		if err == nil { // hit
+			indexes[i] = methodsets.Decode(data)
+			return false
+		} else if err != filecache.ErrNotFound {
+			event.Error(ctx, "reading methodsets from filecache", err)
 		}
+		return true
 	}
-	return indexes, err
+	post := func(i int, pkg *Package) {
+		indexes[i] = pkg.pkg.methodsets
+	}
+	return indexes, s.forEachPackage(ctx, ids, pre, post)
 }
 
 func (s *snapshot) MetadataForFile(ctx context.Context, uri span.URI) ([]*source.Metadata, error) {
@@ -758,14 +811,24 @@ func (s *snapshot) MetadataForFile(ctx context.Context, uri span.URI) ([]*source
 		s.unloadableFiles[uri] = struct{}{}
 	}
 
-	// Sort packages "narrowest" to "widest" (in practice: non-tests before tests).
+	// Sort packages "narrowest" to "widest" (in practice:
+	// non-tests before tests), and regular packages before
+	// their intermediate test variants (which have the same
+	// files but different imports).
 	sort.Slice(metas, func(i, j int) bool {
-		return len(metas[i].CompiledGoFiles) < len(metas[j].CompiledGoFiles)
+		x, y := metas[i], metas[j]
+		xfiles, yfiles := len(x.CompiledGoFiles), len(y.CompiledGoFiles)
+		if xfiles != yfiles {
+			return xfiles < yfiles
+		}
+		return boolLess(x.IsIntermediateTestVariant(), y.IsIntermediateTestVariant())
 	})
 
 	return metas, nil
 }
 
+func boolLess(x, y bool) bool { return !x && y } // false < true
+
 func (s *snapshot) ReverseDependencies(ctx context.Context, id PackageID, transitive bool) (map[PackageID]*source.Metadata, error) {
 	if err := s.awaitLoaded(ctx); err != nil {
 		return nil, err
@@ -796,21 +859,15 @@ func (s *snapshot) ReverseDependencies(ctx context.Context, id PackageID, transi
 	return rdeps, nil
 }
 
-func (s *snapshot) workspaceMetadata() (meta []*source.Metadata) {
-	s.mu.Lock()
-	defer s.mu.Unlock()
-
-	for id := range s.workspacePackages {
-		meta = append(meta, s.meta.metadata[id])
-	}
-	return meta
-}
-
 // -- Active package tracking --
 //
-// We say a package is "active" if any of its files are open. After
-// type-checking we keep active packages in memory. The activePackages
-// peristent map does bookkeeping for the set of active packages.
+// We say a package is "active" if any of its files are open.
+// This is an optimization: the "active" concept is an
+// implementation detail of the cache and is not exposed
+// in the source or Snapshot API.
+// After type-checking we keep active packages in memory.
+// The activePackages persistent map does bookkeeping for
+// the set of active packages.
 
 // getActivePackage returns a the memoized active package for id, if it exists.
 // If id is not active or has not yet been type-checked, it returns nil.
@@ -1037,29 +1094,45 @@ func (s *snapshot) knownFilesInDir(ctx context.Context, dir span.URI) []span.URI
 	return files
 }
 
-func (s *snapshot) ActiveMetadata(ctx context.Context) ([]*source.Metadata, error) {
+func (s *snapshot) WorkspaceMetadata(ctx context.Context) ([]*source.Metadata, error) {
 	if err := s.awaitLoaded(ctx); err != nil {
 		return nil, err
 	}
-	return s.workspaceMetadata(), nil
+
+	s.mu.Lock()
+	defer s.mu.Unlock()
+
+	meta := make([]*source.Metadata, 0, len(s.workspacePackages))
+	for id := range s.workspacePackages {
+		meta = append(meta, s.meta.metadata[id])
+	}
+	return meta, nil
 }
 
 // Symbols extracts and returns symbol information for every file contained in
 // a loaded package. It awaits snapshot loading.
 //
 // TODO(rfindley): move this to the top of cache/symbols.go
-func (s *snapshot) Symbols(ctx context.Context) (map[span.URI][]source.Symbol, error) {
+func (s *snapshot) Symbols(ctx context.Context, workspaceOnly bool) (map[span.URI][]source.Symbol, error) {
 	if err := s.awaitLoaded(ctx); err != nil {
 		return nil, err
 	}
 
-	// Build symbols for all loaded Go files.
-	s.mu.Lock()
-	meta := s.meta
-	s.mu.Unlock()
+	var (
+		meta []*source.Metadata
+		err  error
+	)
+	if workspaceOnly {
+		meta, err = s.WorkspaceMetadata(ctx)
+	} else {
+		meta, err = s.AllMetadata(ctx)
+	}
+	if err != nil {
+		return nil, fmt.Errorf("loading metadata: %v", err)
+	}
 
 	goFiles := make(map[span.URI]struct{})
-	for _, m := range meta.metadata {
+	for _, m := range meta {
 		for _, uri := range m.GoFiles {
 			goFiles[uri] = struct{}{}
 		}
@@ -1194,14 +1267,6 @@ func (s *snapshot) noValidMetadataForURILocked(uri span.URI) bool {
 	return true
 }
 
-func (s *snapshot) isWorkspacePackage(id PackageID) bool {
-	s.mu.Lock()
-	defer s.mu.Unlock()
-
-	_, ok := s.workspacePackages[id]
-	return ok
-}
-
 func (s *snapshot) FindFile(uri span.URI) source.FileHandle {
 	s.view.markKnown(uri)
 
@@ -1212,29 +1277,65 @@ func (s *snapshot) FindFile(uri span.URI) source.FileHandle {
 	return result
 }
 
-// GetFile returns a File for the given URI. If the file is unknown it is added
+// ReadFile returns a File for the given URI. If the file is unknown it is added
 // to the managed set.
 //
-// GetFile succeeds even if the file does not exist. A non-nil error return
+// ReadFile succeeds even if the file does not exist. A non-nil error return
 // indicates some type of internal error, for example if ctx is cancelled.
-func (s *snapshot) GetFile(ctx context.Context, uri span.URI) (source.FileHandle, error) {
+func (s *snapshot) ReadFile(ctx context.Context, uri span.URI) (source.FileHandle, error) {
 	s.mu.Lock()
 	defer s.mu.Unlock()
 
-	return lockedSnapshot{s}.GetFile(ctx, uri)
+	return lockedSnapshot{s}.ReadFile(ctx, uri)
+}
+
+// preloadFiles delegates to the view FileSource to read the requested uris in
+// parallel, without holding the snapshot lock.
+func (s *snapshot) preloadFiles(ctx context.Context, uris []span.URI) {
+	files := make([]source.FileHandle, len(uris))
+	var wg sync.WaitGroup
+	iolimit := make(chan struct{}, 20) // I/O concurrency limiting semaphore
+	for i, uri := range uris {
+		wg.Add(1)
+		iolimit <- struct{}{}
+		go func(i int, uri span.URI) {
+			defer wg.Done()
+			fh, err := s.view.fs.ReadFile(ctx, uri)
+			<-iolimit
+			if err != nil && ctx.Err() == nil {
+				event.Error(ctx, fmt.Sprintf("reading %s", uri), err)
+				return
+			}
+			files[i] = fh
+		}(i, uri)
+	}
+	wg.Wait()
+
+	s.mu.Lock()
+	defer s.mu.Unlock()
+
+	for i, fh := range files {
+		if fh == nil {
+			continue // error logged above
+		}
+		uri := uris[i]
+		if _, ok := s.files.Get(uri); !ok {
+			s.files.Set(uri, fh)
+		}
+	}
 }
 
 // A lockedSnapshot implements the source.FileSource interface while holding
 // the lock for the wrapped snapshot.
 type lockedSnapshot struct{ wrapped *snapshot }
 
-func (s lockedSnapshot) GetFile(ctx context.Context, uri span.URI) (source.FileHandle, error) {
+func (s lockedSnapshot) ReadFile(ctx context.Context, uri span.URI) (source.FileHandle, error) {
 	s.wrapped.view.markKnown(uri)
 	if fh, ok := s.wrapped.files.Get(uri); ok {
 		return fh, nil
 	}
 
-	fh, err := s.wrapped.view.fs.GetFile(ctx, uri) // read the file
+	fh, err := s.wrapped.view.fs.ReadFile(ctx, uri)
 	if err != nil {
 		return nil, err
 	}
@@ -1272,6 +1373,7 @@ func isFileOpen(fh source.FileHandle) bool {
 	return open
 }
 
+// TODO(rfindley): it would make sense for awaitLoaded to return metadata.
 func (s *snapshot) awaitLoaded(ctx context.Context) error {
 	loadErr := s.awaitLoadedAllErrors(ctx)
 
@@ -1283,7 +1385,7 @@ func (s *snapshot) awaitLoaded(ctx context.Context) error {
 	return nil
 }
 
-func (s *snapshot) GetCriticalError(ctx context.Context) *source.CriticalError {
+func (s *snapshot) CriticalError(ctx context.Context) *source.CriticalError {
 	// If we couldn't compute workspace mod files, then the load below is
 	// invalid.
 	//
@@ -1300,7 +1402,7 @@ func (s *snapshot) GetCriticalError(ctx context.Context) *source.CriticalError {
 	// Even if packages didn't fail to load, we still may want to show
 	// additional warnings.
 	if loadErr == nil {
-		active, _ := s.ActiveMetadata(ctx)
+		active, _ := s.WorkspaceMetadata(ctx)
 		if msg := shouldShowAdHocPackagesWarning(s, active); msg != "" {
 			return &source.CriticalError{
 				MainError: errors.New(msg),
@@ -1477,6 +1579,11 @@ func (s *snapshot) reloadWorkspace(ctx context.Context) error {
 	return err
 }
 
+// reloadOrphanedOpenFiles attempts to load a package for each open file that
+// does not yet have an associated package. If loading finishes without being
+// canceled, any files still not contained in a package are marked as unloadable.
+//
+// An error is returned if the load is canceled.
 func (s *snapshot) reloadOrphanedOpenFiles(ctx context.Context) error {
 	// When we load ./... or a package path directly, we may not get packages
 	// that exist only in overlays. As a workaround, we search all of the files
@@ -1510,22 +1617,35 @@ func (s *snapshot) reloadOrphanedOpenFiles(ctx context.Context) error {
 	// mark the failures so we don't bother retrying until the file's
 	// content changes.
 	//
-	// TODO(rstambler): This may be an overestimate if the load stopped
-	// early for an unrelated errors. Add a fallback?
+	// TODO(rfindley): is it possible that the load stopped early for an
+	// unrelated errors? If so, add a fallback?
 	//
 	// Check for context cancellation so that we don't incorrectly mark files
 	// as unloadable, but don't return before setting all workspace packages.
-	if ctx.Err() == nil && err != nil {
-		event.Error(ctx, "reloadOrphanedFiles: failed to load", err, tag.Query.Of(scopes))
-		s.mu.Lock()
-		for _, scope := range scopes {
-			uri := span.URI(scope.(fileLoadScope))
-			if s.noValidMetadataForURILocked(uri) {
-				s.unloadableFiles[uri] = struct{}{}
-			}
-		}
-		s.mu.Unlock()
+	if ctx.Err() != nil {
+		return ctx.Err()
 	}
+
+	// If the context was not canceled, we assume that the result of loading
+	// packages is deterministic (though as we saw in golang/go#59318, it may not
+	// be in the presence of bugs). Marking all unloaded files as unloadable here
+	// prevents us from falling into recursive reloading where we only make a bit
+	// of progress each time.
+	s.mu.Lock()
+	for _, scope := range scopes {
+		// TODO(rfindley): instead of locking here, we should have load return the
+		// metadata graph that resulted from loading.
+		uri := span.URI(scope.(fileLoadScope))
+		if s.noValidMetadataForURILocked(uri) {
+			s.unloadableFiles[uri] = struct{}{}
+		}
+	}
+	s.mu.Unlock()
+
+	if err != nil && !errors.Is(err, errNoPackages) {
+		event.Error(ctx, "reloadOrphanedFiles: failed to load", err, tag.Query.Of(scopes))
+	}
+
 	return nil
 }
 
@@ -1587,15 +1707,15 @@ type unappliedChanges struct {
 	changes          map[span.URI]*fileChange
 }
 
-func (ac *unappliedChanges) GetFile(ctx context.Context, uri span.URI) (source.FileHandle, error) {
+func (ac *unappliedChanges) ReadFile(ctx context.Context, uri span.URI) (source.FileHandle, error) {
 	if c, ok := ac.changes[uri]; ok {
 		return c.fileHandle, nil
 	}
-	return ac.originalSnapshot.GetFile(ctx, uri)
+	return ac.originalSnapshot.ReadFile(ctx, uri)
 }
 
 func (s *snapshot) clone(ctx, bgCtx context.Context, changes map[span.URI]*fileChange, forceReloadMetadata bool) (*snapshot, func()) {
-	ctx, done := event.Start(ctx, "snapshot.clone")
+	ctx, done := event.Start(ctx, "cache.snapshot.clone")
 	defer done()
 
 	reinit := false
@@ -1681,6 +1801,8 @@ func (s *snapshot) clone(ctx, bgCtx context.Context, changes map[span.URI]*fileC
 		knownSubdirs:         s.knownSubdirs.Clone(),
 		workspaceModFiles:    wsModFiles,
 		workspaceModFilesErr: wsModFilesErr,
+		importGraph:          s.importGraph,
+		pkgIndex:             s.pkgIndex,
 	}
 
 	// The snapshot should be initialized if either s was uninitialized, or we've
@@ -1850,9 +1972,16 @@ func (s *snapshot) clone(ctx, bgCtx context.Context, changes map[span.URI]*fileC
 		addRevDeps(id, invalidateMetadata)
 	}
 
-	// Delete invalidated package type information.
-	for id := range idsToInvalidate {
-		result.packages.Delete(id)
+	// Invalidated package information.
+	for id, invalidateMetadata := range idsToInvalidate {
+		if _, ok := directIDs[id]; ok || invalidateMetadata {
+			result.packages.Delete(id)
+		} else {
+			if entry, hit := result.packages.Get(id); hit {
+				ph := entry.(*packageHandle).clone(false)
+				result.packages.Set(id, ph, nil)
+			}
+		}
 		result.activePackages.Delete(id)
 	}
 
@@ -1952,7 +2081,6 @@ func (s *snapshot) clone(ctx, bgCtx context.Context, changes map[span.URI]*fileC
 	if workspaceModeChanged {
 		result.workspacePackages = map[PackageID]PackagePath{}
 	}
-	result.dumpWorkspace("clone")
 	return result, release
 }
 
@@ -2059,9 +2187,10 @@ func metadataChanges(ctx context.Context, lockedSnapshot *snapshot, oldFH, newFH
 		return false, false, false
 	}
 
+	fset := token.NewFileSet()
 	// Parse headers to compare package names and imports.
-	oldHeads, _, oldErr := lockedSnapshot.parseCache.parseFiles(ctx, source.ParseHeader, oldFH)
-	newHeads, _, newErr := lockedSnapshot.parseCache.parseFiles(ctx, source.ParseHeader, newFH)
+	oldHeads, oldErr := lockedSnapshot.parseCache.parseFiles(ctx, fset, source.ParseHeader, oldFH)
+	newHeads, newErr := lockedSnapshot.parseCache.parseFiles(ctx, fset, source.ParseHeader, newFH)
 
 	if oldErr != nil || newErr != nil {
 		// TODO(rfindley): we can get here if newFH does not exist. There is
@@ -2112,8 +2241,8 @@ func metadataChanges(ctx context.Context, lockedSnapshot *snapshot, oldFH, newFH
 	// Note: if this affects performance we can probably avoid parsing in the
 	// common case by first scanning the source for potential comments.
 	if !invalidate {
-		origFulls, _, oldErr := lockedSnapshot.parseCache.parseFiles(ctx, source.ParseFull, oldFH)
-		newFulls, _, newErr := lockedSnapshot.parseCache.parseFiles(ctx, source.ParseFull, newFH)
+		origFulls, oldErr := lockedSnapshot.parseCache.parseFiles(ctx, fset, source.ParseFull, oldFH)
+		newFulls, newErr := lockedSnapshot.parseCache.parseFiles(ctx, fset, source.ParseFull, newFH)
 		if oldErr == nil && newErr == nil {
 			invalidate = magicCommentsChanged(origFulls[0].File, newFulls[0].File)
 		} else {
@@ -2191,11 +2320,18 @@ func (s *snapshot) BuiltinFile(ctx context.Context) (*source.ParsedGoFile, error
 		return nil, fmt.Errorf("no builtin package for view %s", s.view.name)
 	}
 
-	fh, err := s.GetFile(ctx, builtin)
+	fh, err := s.ReadFile(ctx, builtin)
 	if err != nil {
 		return nil, err
 	}
-	return s.ParseGo(ctx, fh, source.ParseFull)
+	// For the builtin file only, we need syntactic object resolution
+	// (since we can't type check).
+	mode := source.ParseFull &^ source.SkipObjectResolution
+	pgfs, err := s.parseCache.parseFiles(ctx, token.NewFileSet(), mode, fh)
+	if err != nil {
+		return nil, err
+	}
+	return pgfs[0], nil
 }
 
 func (s *snapshot) IsBuiltin(ctx context.Context, uri span.URI) bool {
