@@ -577,6 +577,7 @@ func (v *View) shutdown() {
 
 	v.snapshotMu.Lock()
 	if v.snapshot != nil {
+		v.snapshot.cancel()
 		v.releaseSnapshot()
 		v.destroy(v.snapshot, "View.shutdown")
 		v.snapshot = nil
@@ -701,6 +702,12 @@ func (s *snapshot) loadWorkspace(ctx context.Context, firstAttempt bool) (loadEr
 
 	if len(s.workspaceModFiles) > 0 {
 		for modURI := range s.workspaceModFiles {
+			// Verify that the modfile is valid before trying to load it.
+			//
+			// TODO(rfindley): now that we no longer need to parse the modfile in
+			// order to load scope, we could move these diagnostics to a more general
+			// location where we diagnose problems with modfiles or the workspace.
+			//
 			// Be careful not to add context cancellation errors as critical module
 			// errors.
 			fh, err := s.ReadFile(ctx, modURI)
@@ -721,8 +728,11 @@ func (s *snapshot) loadWorkspace(ctx context.Context, firstAttempt bool) (loadEr
 				addError(modURI, fmt.Errorf("no module path for %s", modURI))
 				continue
 			}
-			path := parsed.File.Module.Mod.Path
-			scopes = append(scopes, moduleLoadScope(path))
+			moduleDir := filepath.Dir(modURI.Filename())
+			// Previously, we loaded <modulepath>/... for each module path, but that
+			// is actually incorrect when the pattern may match packages in more than
+			// one module. See golang/go#59458 for more details.
+			scopes = append(scopes, moduleLoadScope{dir: moduleDir, modulePath: parsed.File.Module.Mod.Path})
 		}
 	} else {
 		scopes = append(scopes, viewLoadScope("LOAD_VIEW"))
@@ -731,8 +741,10 @@ func (s *snapshot) loadWorkspace(ctx context.Context, firstAttempt bool) (loadEr
 	// If we're loading anything, ensure we also load builtin,
 	// since it provides fake definitions (and documentation)
 	// for types like int that are used everywhere.
+	// ("unsafe" is also needed since its sole GoFiles is
+	// derived from that of "builtin" via a workaround in load.)
 	if len(scopes) > 0 {
-		scopes = append(scopes, packageLoadScope("builtin"))
+		scopes = append(scopes, packageLoadScope("builtin"), packageLoadScope("unsafe"))
 	}
 	loadErr = s.load(ctx, true, scopes...)
 
