@@ -16,16 +16,19 @@ import (
 	"strings"
 	"sync"
 
+	"golang.org/x/tools/gopls/internal/bug"
 	"golang.org/x/tools/gopls/internal/lsp/debug"
 	"golang.org/x/tools/gopls/internal/lsp/protocol"
 	"golang.org/x/tools/gopls/internal/lsp/source"
 	"golang.org/x/tools/gopls/internal/span"
-	"golang.org/x/tools/internal/bug"
 	"golang.org/x/tools/internal/event"
 	"golang.org/x/tools/internal/jsonrpc2"
 )
 
 func (s *Server) initialize(ctx context.Context, params *protocol.ParamInitialize) (*protocol.InitializeResult, error) {
+	ctx, done := event.Start(ctx, "lsp.Server.initialize")
+	defer done()
+
 	s.stateMu.Lock()
 	if s.state >= serverInitializing {
 		defer s.stateMu.Unlock()
@@ -60,17 +63,17 @@ func (s *Server) initialize(ctx context.Context, params *protocol.ParamInitializ
 
 	if options.ShowBugReports {
 		// Report the next bug that occurs on the server.
-		bugCh := bug.Notify()
-		go func() {
-			b := <-bugCh
+		bug.Handle(func(b bug.Bug) {
 			msg := &protocol.ShowMessageParams{
 				Type:    protocol.Error,
 				Message: fmt.Sprintf("A bug occurred on the server: %s\nLocation:%s", b.Description, b.Key),
 			}
-			if err := s.eventuallyShowMessage(context.Background(), msg); err != nil {
-				log.Printf("error showing bug: %v", err)
-			}
-		}()
+			go func() {
+				if err := s.eventuallyShowMessage(context.Background(), msg); err != nil {
+					log.Printf("error showing bug: %v", err)
+				}
+			}()
+		})
 	}
 
 	folders := params.WorkspaceFolders
@@ -106,7 +109,7 @@ func (s *Server) initialize(ctx context.Context, params *protocol.ParamInitializ
 		}
 	}
 	var renameOpts interface{} = true
-	if r := params.Capabilities.TextDocument.Rename; r.PrepareSupport {
+	if r := params.Capabilities.TextDocument.Rename; r != nil && r.PrepareSupport {
 		renameOpts = protocol.RenameOptions{
 			PrepareProvider: r.PrepareSupport,
 		}
@@ -136,55 +139,55 @@ See https://github.com/golang/go/issues/45732 for more information.`,
 
 	return &protocol.InitializeResult{
 		Capabilities: protocol.ServerCapabilities{
-			CallHierarchyProvider: true,
+			CallHierarchyProvider: &protocol.Or_ServerCapabilities_callHierarchyProvider{Value: true},
 			CodeActionProvider:    codeActionProvider,
 			CodeLensProvider:      &protocol.CodeLensOptions{}, // must be non-nil to enable the code lens capability
-			CompletionProvider: protocol.CompletionOptions{
+			CompletionProvider: &protocol.CompletionOptions{
 				TriggerCharacters: []string{"."},
 			},
-			DefinitionProvider:         true,
-			TypeDefinitionProvider:     true,
-			ImplementationProvider:     true,
-			DocumentFormattingProvider: true,
-			DocumentSymbolProvider:     true,
-			WorkspaceSymbolProvider:    true,
-			ExecuteCommandProvider: protocol.ExecuteCommandOptions{
-				Commands: options.SupportedCommands,
+			DefinitionProvider:         &protocol.Or_ServerCapabilities_definitionProvider{Value: true},
+			TypeDefinitionProvider:     &protocol.Or_ServerCapabilities_typeDefinitionProvider{Value: true},
+			ImplementationProvider:     &protocol.Or_ServerCapabilities_implementationProvider{Value: true},
+			DocumentFormattingProvider: &protocol.Or_ServerCapabilities_documentFormattingProvider{Value: true},
+			DocumentSymbolProvider:     &protocol.Or_ServerCapabilities_documentSymbolProvider{Value: true},
+			WorkspaceSymbolProvider:    &protocol.Or_ServerCapabilities_workspaceSymbolProvider{Value: true},
+			ExecuteCommandProvider: &protocol.ExecuteCommandOptions{
+				Commands: nonNilSliceString(options.SupportedCommands),
 			},
-			FoldingRangeProvider:      true,
-			HoverProvider:             true,
-			DocumentHighlightProvider: true,
-			DocumentLinkProvider:      protocol.DocumentLinkOptions{},
+			FoldingRangeProvider:      &protocol.Or_ServerCapabilities_foldingRangeProvider{Value: true},
+			HoverProvider:             &protocol.Or_ServerCapabilities_hoverProvider{Value: true},
+			DocumentHighlightProvider: &protocol.Or_ServerCapabilities_documentHighlightProvider{Value: true},
+			DocumentLinkProvider:      &protocol.DocumentLinkOptions{},
 			InlayHintProvider:         protocol.InlayHintOptions{},
-			ReferencesProvider:        true,
+			ReferencesProvider:        &protocol.Or_ServerCapabilities_referencesProvider{Value: true},
 			RenameProvider:            renameOpts,
-			SelectionRangeProvider:    protocol.SelectionRangeRegistrationOptions{},
+			SelectionRangeProvider:    &protocol.Or_ServerCapabilities_selectionRangeProvider{Value: true},
 			SemanticTokensProvider: protocol.SemanticTokensOptions{
-				Range: true,
-				Full:  true,
+				Range: &protocol.Or_SemanticTokensOptions_range{Value: true},
+				Full:  &protocol.Or_SemanticTokensOptions_full{Value: true},
 				Legend: protocol.SemanticTokensLegend{
-					TokenTypes:     s.session.Options().SemanticTypes,
-					TokenModifiers: s.session.Options().SemanticMods,
+					TokenTypes:     nonNilSliceString(s.session.Options().SemanticTypes),
+					TokenModifiers: nonNilSliceString(s.session.Options().SemanticMods),
 				},
 			},
-			SignatureHelpProvider: protocol.SignatureHelpOptions{
+			SignatureHelpProvider: &protocol.SignatureHelpOptions{
 				TriggerCharacters: []string{"(", ","},
 			},
 			TextDocumentSync: &protocol.TextDocumentSyncOptions{
 				Change:    protocol.Incremental,
 				OpenClose: true,
-				Save: protocol.SaveOptions{
+				Save: &protocol.SaveOptions{
 					IncludeText: false,
 				},
 			},
-			Workspace: protocol.Workspace6Gn{
-				WorkspaceFolders: protocol.WorkspaceFolders5Gn{
+			Workspace: &protocol.Workspace6Gn{
+				WorkspaceFolders: &protocol.WorkspaceFolders5Gn{
 					Supported:           true,
 					ChangeNotifications: "workspace/didChangeWorkspaceFolders",
 				},
 			},
 		},
-		ServerInfo: protocol.PServerInfoMsg_initialize{
+		ServerInfo: &protocol.PServerInfoMsg_initialize{
 			Name:    "gopls",
 			Version: string(goplsVersion),
 		},
@@ -192,6 +195,9 @@ See https://github.com/golang/go/issues/45732 for more information.`,
 }
 
 func (s *Server) initialized(ctx context.Context, params *protocol.InitializedParams) error {
+	ctx, done := event.Start(ctx, "lsp.Server.initialized")
+	defer done()
+
 	s.stateMu.Lock()
 	if s.state >= serverInitialized {
 		defer s.stateMu.Unlock()
@@ -356,7 +362,7 @@ func (s *Server) addFolders(ctx context.Context, folders []protocol.WorkspaceFol
 		// Diagnose the newly created view asynchronously.
 		ndiagnose.Add(1)
 		go func() {
-			s.diagnoseDetached(snapshot)
+			s.diagnoseSnapshot(snapshot, nil, false, 0)
 			<-initialized
 			release()
 			ndiagnose.Done()
@@ -446,11 +452,12 @@ func (s *Server) registerWatchedDirectoriesLocked(ctx context.Context, patterns
 	for k := range s.watchedGlobPatterns {
 		delete(s.watchedGlobPatterns, k)
 	}
-	var watchers []protocol.FileSystemWatcher
+	watchers := []protocol.FileSystemWatcher{} // must be a slice
+	val := protocol.WatchChange | protocol.WatchDelete | protocol.WatchCreate
 	for pattern := range patterns {
 		watchers = append(watchers, protocol.FileSystemWatcher{
 			GlobPattern: pattern,
-			Kind:        uint32(protocol.WatchChange + protocol.WatchDelete + protocol.WatchCreate),
+			Kind:        &val,
 		})
 	}
 
@@ -568,7 +575,7 @@ func (s *Server) beginFileRequest(ctx context.Context, pURI protocol.DocumentURI
 	if err != nil {
 		return nil, nil, false, func() {}, err
 	}
-	fh, err := snapshot.GetFile(ctx, uri)
+	fh, err := snapshot.ReadFile(ctx, uri)
 	if err != nil {
 		release()
 		return nil, nil, false, func() {}, err
@@ -584,6 +591,9 @@ func (s *Server) beginFileRequest(ctx context.Context, pURI protocol.DocumentURI
 // shutdown implements the 'shutdown' LSP handler. It releases resources
 // associated with the server and waits for all ongoing work to complete.
 func (s *Server) shutdown(ctx context.Context) error {
+	ctx, done := event.Start(ctx, "lsp.Server.shutdown")
+	defer done()
+
 	s.stateMu.Lock()
 	defer s.stateMu.Unlock()
 	if s.state < serverInitialized {
@@ -603,6 +613,9 @@ func (s *Server) shutdown(ctx context.Context) error {
 }
 
 func (s *Server) exit(ctx context.Context) error {
+	ctx, done := event.Start(ctx, "lsp.Server.exit")
+	defer done()
+
 	s.stateMu.Lock()
 	defer s.stateMu.Unlock()
 
@@ -612,7 +625,35 @@ func (s *Server) exit(ctx context.Context) error {
 		// TODO: We should be able to do better than this.
 		os.Exit(1)
 	}
-	// we don't terminate the process on a normal exit, we just allow it to
+	// We don't terminate the process on a normal exit, we just allow it to
 	// close naturally if needed after the connection is closed.
 	return nil
 }
+
+// TODO: when we can assume go1.18, replace with generic
+// (after retiring support for go1.17)
+func nonNilSliceString(x []string) []string {
+	if x == nil {
+		return []string{}
+	}
+	return x
+}
+func nonNilSliceTextEdit(x []protocol.TextEdit) []protocol.TextEdit {
+	if x == nil {
+		return []protocol.TextEdit{}
+	}
+
+	return x
+}
+func nonNilSliceCompletionItemTag(x []protocol.CompletionItemTag) []protocol.CompletionItemTag {
+	if x == nil {
+		return []protocol.CompletionItemTag{}
+	}
+	return x
+}
+func emptySliceDiagnosticTag(x []protocol.DiagnosticTag) []protocol.DiagnosticTag {
+	if x == nil {
+		return []protocol.DiagnosticTag{}
+	}
+	return x
+}
