@@ -7,7 +7,6 @@
 package completion
 
 import (
-	"bytes"
 	"context"
 	"fmt"
 	"go/ast"
@@ -27,6 +26,7 @@ import (
 
 	"golang.org/x/sync/errgroup"
 	"golang.org/x/tools/go/ast/astutil"
+	goplsastutil "golang.org/x/tools/gopls/internal/astutil"
 	"golang.org/x/tools/gopls/internal/lsp/protocol"
 	"golang.org/x/tools/gopls/internal/lsp/safetoken"
 	"golang.org/x/tools/gopls/internal/lsp/snippet"
@@ -446,7 +446,7 @@ func Completion(ctx context.Context, snapshot source.Snapshot, fh source.FileHan
 
 	startTime := time.Now()
 
-	pkg, pgf, err := source.PackageForFile(ctx, snapshot, fh.URI(), source.NarrowestPackage)
+	pkg, pgf, err := source.NarrowestPackageForFile(ctx, snapshot, fh.URI())
 	if err != nil || pgf.File.Package == token.NoPos {
 		// If we can't parse this file or find position for the package
 		// keyword, it may be missing a package declaration. Try offering
@@ -658,7 +658,11 @@ func (c *completer) collectCompletions(ctx context.Context) error {
 	//   recv.‸(arg)
 	case *ast.TypeAssertExpr:
 		// Create a fake selector expression.
-		return c.selector(ctx, &ast.SelectorExpr{X: n.X})
+		//
+		// The name "_" is the convention used by go/parser to represent phantom
+		// selectors.
+		sel := &ast.Ident{NamePos: n.X.End() + token.Pos(len(".")), Name: "_"}
+		return c.selector(ctx, &ast.SelectorExpr{X: n.X, Sel: sel})
 	case *ast.SelectorExpr:
 		return c.selector(ctx, n)
 	// At the file scope, only keywords are allowed.
@@ -912,14 +916,14 @@ func (c *completer) populateCommentCompletions(ctx context.Context, comment *ast
 	// comment itself.
 	c.opts.documentation = false
 
-	commentLine := file.Line(comment.End())
+	commentLine := safetoken.Line(file, comment.End())
 
 	// comment is valid, set surrounding as word boundaries around cursor
 	c.setSurroundingForComment(comment)
 
 	// Using the next line pos, grab and parse the exported symbol on that line
 	for _, n := range c.file.Decls {
-		declLine := file.Line(n.Pos())
+		declLine := safetoken.Line(file, n.Pos())
 		// if the comment is not in, directly above or on the same line as a declaration
 		if declLine != commentLine && declLine != commentLine+1 &&
 			!(n.Pos() <= comment.Pos() && comment.End() <= n.End()) {
@@ -1134,7 +1138,6 @@ func (c *completer) selector(ctx context.Context, sel *ast.SelectorExpr) error {
 		// Imported declaration with missing type information.
 		// Fall through to shallow completion of unimported package members.
 		// Match candidate packages by path.
-		// TODO(adonovan): simplify by merging with else case and matching on name only?
 		filter = func(m *source.Metadata) bool {
 			return strings.TrimPrefix(string(m.PkgPath), "vendor/") == imp.Path()
 		}
@@ -1171,7 +1174,8 @@ func (c *completer) selector(ctx context.Context, sel *ast.SelectorExpr) error {
 	// not assume global Pos/Object realms and then use export
 	// data instead of the quick parse approach taken here.
 
-	// First, we search among packages in the workspace.
+	// First, we search among packages in the forward transitive
+	// closure of the workspace.
 	// We'll use a fast parse to extract package members
 	// from those that match the name/path criterion.
 	all, err := c.snapshot.AllMetadata(ctx)
@@ -1212,11 +1216,11 @@ func (c *completer) selector(ctx context.Context, sel *ast.SelectorExpr) error {
 			return nil
 		}
 
-		fh, err := c.snapshot.GetFile(ctx, uri)
+		fh, err := c.snapshot.ReadFile(ctx, uri)
 		if err != nil {
 			return err
 		}
-		content, err := fh.Read()
+		content, err := fh.Content()
 		if err != nil {
 			return err
 		}
@@ -1607,7 +1611,7 @@ func (c *completer) unimportedPackages(ctx context.Context, seen map[string]stru
 
 	count := 0
 
-	// Search packages across the entire workspace.
+	// Search the forward transitive closure of the workspace.
 	all, err := c.snapshot.AllMetadata(ctx)
 	if err != nil {
 		return err
@@ -3173,7 +3177,7 @@ func isSlice(obj types.Object) bool {
 // quick partial parse. fn is non-nil only for function declarations.
 // The AST position information is garbage.
 func forEachPackageMember(content []byte, f func(tok token.Token, id *ast.Ident, fn *ast.FuncDecl)) {
-	purged := purgeFuncBodies(content)
+	purged := goplsastutil.PurgeFuncBodies(content)
 	file, _ := parser.ParseFile(token.NewFileSet(), "", purged, 0)
 	for _, decl := range file.Decls {
 		switch decl := decl.(type) {
@@ -3195,58 +3199,3 @@ func forEachPackageMember(content []byte, f func(tok token.Token, id *ast.Ident,
 		}
 	}
 }
-
-// purgeFuncBodies returns a copy of src in which the contents of each
-// outermost {...} region except struct and interface types have been
-// deleted. It does not preserve newlines. This reduces the amount of
-// work required to parse the top-level declarations.
-func purgeFuncBodies(src []byte) []byte {
-	// Destroy the content of any {...}-bracketed regions that are
-	// not immediately preceded by a "struct" or "interface"
-	// token.  That includes function bodies, composite literals,
-	// switch/select bodies, and all blocks of statements.
-	// This will lead to non-void functions that don't have return
-	// statements, which of course is a type error, but that's ok.
-
-	var out bytes.Buffer
-	file := token.NewFileSet().AddFile("", -1, len(src))
-	var sc scanner.Scanner
-	sc.Init(file, src, nil, 0)
-	var prev token.Token
-	var cursor int         // last consumed src offset
-	var braces []token.Pos // stack of unclosed braces or -1 for struct/interface type
-	for {
-		pos, tok, _ := sc.Scan()
-		if tok == token.EOF {
-			break
-		}
-		switch tok {
-		case token.COMMENT:
-			// TODO(adonovan): opt: skip, to save an estimated 20% of time.
-
-		case token.LBRACE:
-			if prev == token.STRUCT || prev == token.INTERFACE {
-				pos = -1
-			}
-			braces = append(braces, pos)
-
-		case token.RBRACE:
-			if last := len(braces) - 1; last >= 0 {
-				top := braces[last]
-				braces = braces[:last]
-				if top < 0 {
-					// struct/interface type: leave alone
-				} else if len(braces) == 0 { // toplevel only
-					// Delete {...} body.
-					start, _ := safetoken.Offset(file, top)
-					end, _ := safetoken.Offset(file, pos)
-					out.Write(src[cursor : start+len("{")])
-					cursor = end
-				}
-			}
-		}
-		prev = tok
-	}
-	out.Write(src[cursor:])
-	return out.Bytes()
-}
